@@ -5,6 +5,7 @@ require('tmpdir')
 require('fileutils')
 
 class CompileCacheKeyFormatTest < Minitest::Test
+  FILE = File.expand_path(__FILE__)
   include(TmpdirHelper)
 
   R = {
@@ -18,16 +19,16 @@ class CompileCacheKeyFormatTest < Minitest::Test
   }
 
   def test_key_version
-    key = cache_key_for_file(__FILE__)
-    exp = [2].pack("L")
+    key = cache_key_for_file(FILE)
+    exp = [3].pack("L")
     assert_equal(exp, key[R[:version]])
   end
 
   def test_key_compile_option_stable
-    k1 = cache_key_for_file(__FILE__)
-    k2 = cache_key_for_file(__FILE__)
+    k1 = cache_key_for_file(FILE)
+    k2 = cache_key_for_file(FILE)
     RubyVM::InstructionSequence.compile_option = { tailcall_optimization: true }
-    k3 = cache_key_for_file(__FILE__)
+    k3 = cache_key_for_file(FILE)
     assert_equal(k1[R[:compile_option]], k2[R[:compile_option]])
     refute_equal(k1[R[:compile_option]], k3[R[:compile_option]])
   ensure
@@ -35,7 +36,7 @@ class CompileCacheKeyFormatTest < Minitest::Test
   end
 
   def test_key_ruby_revision
-    key = cache_key_for_file(__FILE__)
+    key = cache_key_for_file(FILE)
     exp = if RUBY_REVISION.is_a?(String)
       [Help.fnv1a_64(RUBY_REVISION) >> 32].pack("L")
     else
@@ -45,39 +46,49 @@ class CompileCacheKeyFormatTest < Minitest::Test
   end
 
   def test_key_size
-    key = cache_key_for_file(__FILE__)
-    exp = File.size(__FILE__)
+    key = cache_key_for_file(FILE)
+    exp = File.size(FILE)
     act = key[R[:size]].unpack("Q")[0]
     assert_equal(exp, act)
   end
 
   def test_key_mtime
-    key = cache_key_for_file(__FILE__)
-    exp = File.mtime(__FILE__).to_i
+    key = cache_key_for_file(FILE)
+    exp = File.mtime(FILE).to_i
     act = key[R[:mtime]].unpack("Q")[0]
     assert_equal(exp, act)
   end
 
   def test_fetch
-    actual = Bootsnap::CompileCache::Native.fetch(@tmp_dir, '/dev/null', TestHandler)
-    assert_equal('NEATO /DEV/NULL', actual)
-    data = File.read("#{@tmp_dir}/8c/d2d180bbd995df")
-    assert_equal("neato /dev/null", data.force_encoding(Encoding::BINARY)[64..-1])
-    actual = Bootsnap::CompileCache::Native.fetch(@tmp_dir, '/dev/null', TestHandler)
-    assert_equal('NEATO /DEV/NULL', actual)
+    if RbConfig::CONFIG['host_os'] =~ /mswin|mingw|cygwin/
+      target = 'NUL'
+      expected_file = "#{@tmp_dir}/36/9eba19c29ffe00"
+    else
+      target = '/dev/null'
+      expected_file = "#{@tmp_dir}/8c/d2d180bbd995df"
+    end
+
+    actual = Bootsnap::CompileCache::Native.fetch(@tmp_dir, target, TestHandler, nil)
+    assert_equal("NEATO #{target.upcase}", actual)
+
+    data = File.read(expected_file)
+    assert_equal("neato #{target}", data.force_encoding(Encoding::BINARY)[64..-1])
+
+    actual = Bootsnap::CompileCache::Native.fetch(@tmp_dir, target, TestHandler, nil)
+    assert_equal("NEATO #{target.upcase}", actual)
   end
 
   def test_unexistent_fetch
     assert_raises(Errno::ENOENT) do
-      Bootsnap::CompileCache::Native.fetch(@tmp_dir, '123', Bootsnap::CompileCache::ISeq)
+      Bootsnap::CompileCache::Native.fetch(@tmp_dir, '123', Bootsnap::CompileCache::ISeq, nil)
     end
   end
 
   private
 
   def cache_key_for_file(file)
-    Bootsnap::CompileCache::Native.fetch(@tmp_dir, file, TestHandler)
-    data = File.read(Help.cache_path(@tmp_dir, file))
-    Help.binary(data[0..31])
+    Bootsnap::CompileCache::Native.fetch(@tmp_dir, file, TestHandler, nil)
+    data = File.binread(Help.cache_path(@tmp_dir, file))
+    data.byteslice(0..31)
   end
 end
