@@ -24,11 +24,24 @@ class CompileCacheTest < Minitest::Test
   end
 
   def test_no_write_permission_to_cache
-    path = Help.set_file('a.rb', 'a = 3', 100)
-    folder = File.dirname(Help.cache_path(@tmp_dir, path))
-    FileUtils.mkdir_p(folder)
-    FileUtils.chmod(0400, folder)
-    assert_raises(Bootsnap::CompileCache::PermissionError) { load(path) }
+    if RbConfig::CONFIG['host_os'] =~ /mswin|mingw|cygwin/
+      # Always pass this test on Windows because directories aren't read, only
+      # listed. You can restrict the ability to list directory contents on
+      # Windows or you can set ACLS on a folder such that it is not allowed to
+      # list contents.
+      #
+      # Since we can't read directories on windows, this specific test doesn't
+      # make sense. In addtion we test read-only files in
+      # `test_can_open_read_only_cache` so we are covered testing reading
+      # read-only files.
+      pass
+    else
+      path = Help.set_file('a.rb', 'a = 3', 100)
+      folder = File.dirname(Help.cache_path(@tmp_dir, path))
+      FileUtils.mkdir_p(folder)
+      FileUtils.chmod(0400, folder)
+      load(path)
+    end
   end
 
   def test_can_open_read_only_cache
@@ -106,4 +119,46 @@ class CompileCacheTest < Minitest::Test
     load(path)
     assert(File.size(cp) > 32) # cache was overwritten
   end
+
+  def test_instrumentation_hit
+    path = Help.set_file('a.rb', 'a = a = 3', 100)
+    load(path)
+
+    calls = []
+    Bootsnap.instrumentation = ->(event, path) { calls << [event, path] }
+
+    load(path)
+
+    assert_equal [], calls
+  ensure
+    Bootsnap.instrumentation = nil
+  end
+
+  def test_instrumentation_miss
+    path = Help.set_file('a.rb', 'a = a = 3', 100)
+
+    calls = []
+    Bootsnap.instrumentation = ->(event, path) { calls << [event, path] }
+
+    load(path)
+
+    assert_equal [[:miss, 'a.rb']], calls
+  ensure
+    Bootsnap.instrumentation = nil
+  end
+
+  def test_instrumentation_stale
+    path = Help.set_file('a.rb', 'a = a = 3', 100)
+    load(path)
+    path = Help.set_file('a.rb', 'a = a = 4', 101)
+
+    calls = []
+    Bootsnap.instrumentation = ->(event, path) { calls << [event, path] }
+
+    load(path)
+
+    assert_equal [[:stale, 'a.rb']], calls
+  ensure
+    Bootsnap.instrumentation = nil
+  end
 end
