@@ -17,18 +17,29 @@ module Bootsnap
         @observer.expects(:push_paths).with(@arr, 'b', 'c')
         @arr.push('b', 'c')
 
-        @observer.expects(:unshift_paths).with(@arr, 'd', 'e')
-        @arr.unshift('d', 'e')
+        @observer.expects(:push_paths).with(@arr, 'd', 'e')
+        @arr.append('d', 'e')
 
-        @observer.expects(:push_paths).with(@arr, 'f', 'g')
-        @arr.concat(%w(f g))
+        @observer.expects(:unshift_paths).with(@arr, 'f', 'g')
+        @arr.unshift('f', 'g')
+
+        @observer.expects(:push_paths).with(@arr, 'h', 'i')
+        @arr.concat(%w(h i))
+
+        @observer.expects(:unshift_paths).with(@arr, 'j', 'k')
+        @arr.prepend('j', 'k')
+      end
+
+      def test_reinitializes_on_aggressive_modifications
+        @observer.expects(:push_paths).with(@arr, 'a', 'b', 'c')
+        @arr.push('a', 'b', 'c')
 
         @observer.expects(:reinitialize).times(4)
         @arr.delete(3)
         @arr.compact!
         @arr.map!(&:upcase)
-        assert_equal('G', @arr.pop)
-        assert_equal(%w(D E A B C F), @arr)
+        assert_equal('C', @arr.pop)
+        assert_equal(%w(A B), @arr)
       end
 
       def test_register_frozen
